type binop' =
  | Add'
  | Mul'

type expr' =
  | Const' of int
  | Binop' of binop' * expr' * expr'
  | Abs' of (expr' -> expr' option)
  | App' of expr' * expr'

let rec eval' e = match e with
    | Const' i -> Some (Const' i)
    | Binop'(Add', u, v) ->
      let x = eval' v in
      begin match eval' u with
        | ((((None | Some (Binop'(Add', _, _)))
          | Some (Binop'(Mul', _, _)))
          | Some (Abs' _))
          | Some (App'(_, _))) -> None
        | Some (Const' i) ->
          begin match x with
            | ((((None | Some (Binop'(Add', _, _)))
              | Some (Binop'(Mul', _, _)))
              | Some (Abs' _))
              | Some (App'(_, _))) -> None
            | Some (Const' i1) -> Some (Const' (add i i1))
          end
      end
    | Binop'(Mul', u, v) ->
      let x = eval' v in
      begin match eval' u with
        | ((((None | Some (Binop'(Add', _, _)))
          | Some (Binop'(Mul', _, _)))
          | Some (Abs' _))
          | Some (App'(_, _))) -> None
        | Some (Const' i) ->
          begin match x with
            | ((((None | Some (Binop'(Add', _, _)))
              | Some (Binop'(Mul', _, _)))
              | Some (Abs' _))
              | Some (App'(_, _))) -> None
            | Some (Const' i1) -> Some (Const' (mul i i1))
          end
      end
    | Abs' f -> Some (Abs' f)
    | App'(u, v) ->
      begin match eval' u with
        | ((((None | Some (Const' _))
          | Some (Binop'(Add', _, _)))
          | Some (Binop'(Mul', _, _)))
          | Some (App'(_, _))) -> None
        | Some (Abs' f) ->
          begin match eval' v with
            | None -> None
            | Some x -> f x
          end
      end

type expr'' =
  | Const'' of int
  | Binop'' of binop' * expr'' * expr''
  | Abs'' of (value'' -> value'' option)
  | App'' of expr'' * expr''
and value'' =
  | Int'' of int
  | Fun'' of (value'' -> value'' option)

let rec eval'' e = match e with
    | Const'' i -> Some (Int'' i)
    | Binop''(Add', u, v) ->
      let x = eval'' v in
      begin match eval'' u with
        | (None | Some (Fun'' _)) -> None
        | Some (Int'' i) ->
          begin match x with
            | (None | Some (Fun'' _)) -> None
            | Some (Int'' i1) -> Some (Int'' (add i i1))
          end
      end
    | Binop''(Mul', u, v) ->
      let x = eval'' v in
      begin match eval'' u with
        | (None | Some (Fun'' _)) -> None
        | Some (Int'' i) ->
          begin match x with
            | (None | Some (Fun'' _)) -> None
            | Some (Int'' i1) -> Some (Int'' (mul i i1))
          end
      end
    | Abs'' f -> Some (Fun'' f)
    | App''(u, v) ->
      begin match eval'' u with
        | (None | Some (Int'' _)) -> None
        | Some (Fun'' f) ->
          begin match eval'' v with
            | None -> None
            | Some x -> f x
          end
      end
