type location = Location of string * int * int

type expr_loc =
  | Abs' of (expr_loc -> expr_loc) * location
  | App' of expr_loc * expr_loc * location
  | Const' of int * location

type ('a,'err) result =
  | Ok of 'a
  | Error of 'err

let rec _eval e = match e with
    | (Abs'(_, _) | Const'(_, _)) -> Ok e
    | App'(u, v, _) ->
      begin match _eval u with
        | Ok (Abs'(f, _)) -> _eval (f v)
        | ((Ok (App'(_, _, _)) | Ok (Const'(_, _)))
          | Error _) -> Error #4
      end

let rec eval_loc e = match e with
    | (Abs'(_, _) | Const'(_, _)) -> Ok e
    | App'(u, v, loc) ->
      begin match eval_loc u with
        | Ok (Abs'(f, _)) -> eval_loc (f v)
        | (Ok (App'(_, _, _)) | Ok (Const'(_, _))) ->
          Error loc
        | Error x -> Error x
      end

type expr' =
  | Abs' of ((expr' * location) -> (expr' * location))
  | App' of (expr' * location) * (expr' * location)
  | Const' of int

let rec _eval e = match e with
    | ((Abs' _, _) | (Const' _, _)) -> Ok e
    | (App'(u, v), _) ->
      begin match _eval u with
        | Ok (Abs' f, _) -> _eval (f v)
        | ((Ok (App'(_, _), _) | Ok (Const' _, _)) | Error _) ->
          Error #4
      end

let rec eval_loc e = match e with
    | ((Abs' _, _) | (Const' _, _)) -> Ok e
    | (App'(u, v), loc) ->
      begin match eval_loc u with
        | Ok (Abs' f, _) -> eval_loc (f v)
        | (Ok (App'(_, _), _) | Ok (Const' _, _)) ->
          Error loc
        | Error x -> Error x
      end
