
# 95 "expr_smaller.tex"
type ornament expr_value : expr => value with
  | Abs(x, e) => VAbs(x, e)
  | Tup(e) => VTup(e) when e : expr_value
  | Var(x) => ~
  | App(e1,e2) => ~
  | Proj(e) => ~

# 106 "expr_smaller.tex"
let eval' = lifting eval : (id * expr_value) list -> expr -> expr_value option

# 138 "expr_smaller.tex"
type value' =
  | VClos' of id * (id * value') list * expr
  | VTup' of value'

# 146 "expr_smaller.tex"
type ornament value_value' : value => value' with
  | VAbs(x, e) => VClos'(x, _, e)
  | VTup(v) => VTup'(v) when v : value_value'

# 157 "expr_smaller.tex"
let eval'' = lifting eval' with
  | ornament * <- value_value', @id
  | * <- env

# 186 "expr_smaller.tex"
let eval'' = lifting eval' with | ornament * <- value_value', @id

# 216 "expr_smaller.tex"
let eval'' = lifting eval' with
  | ornament * <- value_value', @id
  | #37 <- env

# 295 "expr_smaller.tex"
type expr' =
  | Var' of id
  | Abs' of id * expr'
  | Closure' of id * (id * expr') list * expr'
  | App' of expr' * expr'
  | Tup' of expr'
  | Proj' of expr'

# 304 "expr_smaller.tex"
type ornament expr_expr' : expr => expr' with
  | Var(x) => Var'(x)
  | Abs(x, e) => (Abs'(x, e) | Closure'(x,_,e)) when e : expr_expr'
  | App(e1, e2) => App'(e1, e2) when e1 e2 : expr_expr'
  | Tup(e) => Tup'(e) when e : expr_expr'
  | Proj(e) => Proj'(e) when e : expr_expr'

# 315 "expr_smaller.tex"
let eval' = lifting eval : (id * expr_expr') list -> expr_expr' -> expr_expr' option with
  | ornament * <- @id
  | #51 <- Right env

# 350 "expr_smaller.tex"
let rec eval' env e = match e with
    | Var' x -> assoc x env
    | Abs'(x, e) -> Some (Closure'(x, env, e))
    | Closure'(x, _, e) -> Some (Closure'(x, env, e))
    | App'(e1, e2) ->
      bind (eval' env e1)
      (fun v ->
          match v with
            | Var' _ -> fail ()
            | Abs'(x, e) ->
              bind (eval' env e2)
                (fun v -> eval' (Cons((x, v), env)) e)
            | Closure'(x, closure_env, e) ->
              bind (eval' env e2)
                (fun v -> eval' (Cons((x, v), closure_env (* was env *))) e)
            | App'(_, _) -> fail ()
            | Tup' _ -> None
            | Proj' _ -> fail ())
    | Tup' e ->
      bind (eval' env e) (fun v -> Some (Tup' v))
    | Proj' e ->
      bind (eval' env e)
        (fun v ->
            match v with
              | Var' _ -> fail ()
              | (Abs'(_, _) |Closure'(_, _, _)) -> None
              | App'(_, _) -> fail ()
              | Tup' e -> Some e
              | Proj' _ -> fail ())

# 385 "expr_smaller.tex"
type ornament expr'_expr : expr' => expr with
  | Var'(x) => Var(x)
  | Abs'(x,e) => Abs(x,e) when e : expr'_expr
  | App'(e1,e2) => App(e1,e2) when e1 e2 : expr'_expr
  | Tup'(e) => Tup(e) when e : expr'_expr
  | Proj'(e) => Proj(e) when e : expr'_expr
  | Closure'(x,env,e) => ~ when e : expr'_expr
type ornament expr'_value' : expr' => value' with
  | Abs'(x, e) => ~ when e : expr'_expr
  | Closure'(x,env,e) => VClos'(x,env,e) when env : (id * expr'_value') list, e : expr'_expr
  | Tup'(e) => VTup'(e) when e : expr'_value'
  | Var'(x) => ~
  | App'(e1,e2) => ~
  | Proj'(e) => ~

# 403 "expr_smaller.tex"
let eval'' = lifting eval' : (id * expr'_value') list -> expr'_expr -> expr'_value' option
