type expr =
  | Var of string
  | Abs of string * expr
  | App of expr * expr
  | Const of int

type ('a,'err) result =
  | Ok of 'a
  | Error of 'err

let rec eval_mes_ env e = match e with
    | Var x ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error #23
      end
    | (Abs(_, _) | Const _) -> Ok e
    | App(u, v) ->
      begin match eval_mes_ env u with
        | ((Ok (Var _) | Ok (App(_, _))) | Ok (Const _)) ->
          Error #6
        | Ok (Abs(x, f)) ->
          begin match eval_mes_ env v with
            | Ok a -> eval_mes_ (Cons((x, a), env)) f
            | Error _ -> Error #9
          end
        | Error _ -> Error #4
      end

let rec eval_mes env e = match e with
    | Var x ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error str__Unbound_variable
      end
    | (Abs(_, _) | Const _) -> Ok e
    | App(u, v) ->
      begin match eval_mes env u with
        | ((Ok (Var _) | Ok (App(_, _))) | Ok (Const _)) ->
          Error str__Not_a_function
        | Ok (Abs(x, f)) ->
          begin match eval_mes env v with
            | Ok a -> eval_mes (Cons((x, a), env)) f
            | Error mes_v -> Error mes_v
          end
        | Error mes_u -> Error mes_u
      end

type location = Location of string * int * int

type expr' =
  | Var' of string
  | Abs' of string * (expr' * location)
  | App' of (expr' * location) * (expr' * location)
  | Const' of int

type ('a,'err) result' =
  | Ok' of 'a
  | Error' of 'err

let rec _eval_loc env e = match e with
    | (Var' x, _) ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error (str__Unbound_variable, #26)
      end
    | ((Abs'(_, _), _) | (Const' _, _)) -> Ok e
    | (App'(u, v), _) ->
      begin match _eval_loc env u with
        | ((Ok (Var' _, _) | Ok (App'(_, _), _))
          | Ok (Const' _, _)) ->
          Error (str__Not_a_function, #6)
        | Ok (Abs'(x, f), _) ->
          begin match _eval_loc env v with
            | Ok a -> _eval_loc (Cons((x, a), env)) f
            | Error (b, _) -> Error (b, #11)
          end
        | Error (b, _) -> Error (b, #4)
      end

let rec eval_loc env e = match e with
    | (Var' x, _) ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error (str__Unbound_variable, #26)
      end
    | ((Abs'(_, _), _) | (Const' _, _)) -> Ok e
    | (App'(u, v), _) ->
      begin match eval_loc env u with
        | ((Ok (Var' _, loc_naf) | Ok (App'(_, _), loc_naf))
          | Ok (Const' _, loc_naf)) ->
          Error (str__Not_a_function, loc_naf)
        | Ok (Abs'(x, f), _) ->
          begin match eval_loc env v with
            | Ok a -> eval_loc (Cons((x, a), env)) f
            | Error (b, _) -> Error (b, #11)
          end
        | Error (b, loc_u) -> Error (b, loc_u)
      end

let rec _eval_loc env e = match e with
    | (Var' x, _) ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error str__Unbound_variable
      end
    | ((Abs'(_, _), _) | (Const' _, _)) -> Ok e
    | (App'(u, v), _) ->
      begin match _eval_loc env u with
        | ((Ok (Var' _, _) | Ok (App'(_, _), _))
          | Ok (Const' _, _)) -> Error str__Not_a_function
        | Ok (Abs'(x, f), _) ->
          begin match _eval_loc env v with
            | Ok a -> _eval_loc (Cons((x, a), env)) f
            | Error mes_v -> Error mes_v
          end
        | Error mes_u -> Error mes_u
      end

let rec free_vars_loc e = match e with
    | (Var' x, _) -> Cons(x, Nil)
    | (Abs'(x, f), _) -> remove x (free_vars_loc f)
    | (App'(u, v), _) ->
      append (free_vars_loc u) (free_vars_loc v)
    | (Const' _, _) -> Nil
