%{
  open Ast
  module O = Orn.Implicit
(*  module P = Patch.Implicit
  open P *)

  let curfile = Util.curfile

  let make_function args e =
    List.fold_right (fun x ((_,loc) as e) -> Abs(x,e), loc) args e

  let mk_tuple_ty = function
    | [t] -> t
    | l -> TyProduct(l)
%}

%token Match With Type Of Let Rec In Fun And Val
%token Ornament Lifting  Begin End Function
%token <string> Con Ident
%token <int> IntLit
%token Pipe Arrow Equal Lparen Rparen Quote Comma Star Underscore
%token Colon Semicolon
%token EOF Tilde Sharp At RevArrow
%token DoubleArrow When


(* Partial order for application: Ident, Lparen > app
   test:
Let Ident Equal Ident Ident Ident
*)


(* partial order for types : tarrow > Arrow, tarrow > Ident

 test 1 (x -> x -> x => x -> (x -> x)):
  Type Ident Equal Con Of Quote Ident Arrow Quote Ident Arrow Quote Ident EOF
  should lean right

 test 2 (x -> x x => x -> (x x)):
   Type Ident Equal Con Of Ident Arrow Ident Ident EOF
  should lean right
*)
%nonassoc semione
%nonassoc Semicolon

%right tarrow
%right Arrow
%nonassoc Ident
%nonassoc Equal

//%nonassoc Con
//%nonassoc Lparen

//%right app

(* For pattern matching, clauses_empty > Pipe

  testcase:
Let Ident Equal Match Ident With Con Arrow Match Ident With Con Arrow Ident Pipe Con Arrow Ident EOF
*)
%nonassoc clauses_empty
%right Pipe


%start toplevel

%type <Ast.stmt list> toplevel
%type <Id.user_id> ident conid
%type <Ast.pat> pat

%%

%inline
located(X):
  x=X { (x, makepos !curfile $startpos $endpos) }

toplevel:
  | decls = decl* EOF { decls }

decl_l:
  | Sharp Lparen l=ident_plus* Rparen { StPragma(l) }
  | Type ts=separated_list(And,tydecl) { StType(ts) }
  | Let ir=isrec bs=separated_nonempty_list(And,binding)  { StVal(ir,bs) }
//  | Let Rec i=ident a=ident* Equal e=expr { StVal(i,true,make_function a e) }
  | Val i=ident Colon t = typ { StAbstractVal(i,t) }
  | Type Ornament os=separated_list(And,odecl) { StOrnDef(os) }
// We have to disambiguate
  | Let ls=separated_nonempty_list(And,lspec)   {
      O.(StLiftDef(ls))
     }

%inline
lspec:
  | a=ident l=list(ident) Equal Lifting b=ident t=annot? ps=withpatches {
    assert (l = []); (* Avoids a conflict *)
    {lift_from = b; lift_to = a; lift_with = ps; lift_at = t; lift_loc = makepos !curfile $startpos $endpos }
  }

ident_plus:
  | i=ident { i }
  | Lifting { UserId "lifting" }

withpatches:
  | With Pipe? ps=separated_nonempty_list(Pipe,located(patch)) { ps }
  | { [] }

odecl:
  tvs = tyvars i = ident Colon tfrom=typ DoubleArrow tto=typ With
    Pipe? c1=oclause cs=oclauses {
      {orn_name = i; orn_args = tvs; orn_from = tfrom; orn_to = tto; orn_clauses = c1::cs}

   }
annot:
 Colon t=typ { t }

%inline
decl:
 x = located(decl_l) { x }

tydecl:
  | tvs = tyvars i=ident conses=tyconses { TyDef(i,tvs,conses) }

tyconses:
  | Equal Pipe? cs=separated_nonempty_list(Pipe, condecl) { cs }

condecl:
  | i=conid { (i,[]) }
  | i=conid Of ts=separated_nonempty_list(Star, typ2) { (i,ts) }

tyvars:
  | {[]}
  | t = tyvar {[t]}
  | t = delimited(Lparen, separated_nonempty_list(Comma,tyvar), Rparen) {t}

tyvar: Quote i=ident { i }

typ:
  | u = typ Arrow v = typ { TyArrow(u,v) } %prec tarrow
  | l=separated_nonempty_list(Star, typ2) { mk_tuple_ty l }

typ2:
  | Underscore { TyVar(UserId "_") }
  | Lparen u=typ Rparen { u }
  | Lparen Rparen { TyProduct [] }
  | ps=plist(typ,typ2) i=ident { TyCon(i,ps) }
  | Quote i=ident { TyVar i }

%inline
plist(X,Y):
  | { [] }
  | x=Y { [x] }
  | xs=delimited(Lparen,at_least_two(X),Rparen) { xs }

%inline
at_least_two(X):
  | x = X Comma xs = separated_nonempty_list(Comma,X) { x::xs }

binding:
  | i = ident a=ident* Equal v = expr { (i,make_function a v) }

%inline
isrec:
  | { false }
  | Rec { true}

expr_l:
  | e = expr2_con_l { e }
  | Function Pipe? c1=clause cs=clauses {
    let x = UserId "__function__" in
    let loc t = t, makepos !curfile $startpos $endpos in
    Abs(x, loc ( Match( (loc (Var x)), c1::cs ) ) )
  }
  | Match e = expr With Pipe? c1=clause cs=clauses { Match(e,c1::cs) }
  | Let isrec=isrec bs=separated_nonempty_list(And,binding) In e=expr { Let(IsUser, isrec,bs,e) }
  | Fun i=ident Arrow e=expr { Abs(i,e) }

%inline
expr:
  x = located(expr_l) { x }

clauses:
  | { [] } %prec clauses_empty
  | Pipe c=clause cs=clauses { c::cs }

expr2_con_l:
  | c = conid { Con(c,[]) }
  | e = expr2_head_l { e }

expr2_head_l:
  | i = ident { Var(i) }
  | e = expr2_head f = expr2 { App(e,f) } // %prec app
  | c = conid f = expr2 { Con(c,[f]) }
  | Begin e = expr_l End { e }
  | xs=delimited(Lparen,separated_list(Comma,expr),Rparen)
    { mk_tuple (fun (x : expr list) -> Tuple x) (xs : expr list) }

%inline
expr2_head:
  x = located(expr2_head_l) { x }

expr2_l:
  | c = conid  { Con(c,[]) }
  | i = ident { Var(i) }
  | xs=delimited(Lparen,separated_list(Comma,expr),Rparen) { mk_tuple (fun x -> Tuple x) xs }
  | Begin e = expr_l End { e }

%inline
expr2:
  x = located(expr2_l) { x }

pat:
 x = pat_l { (x, makepos "" $startpos $endpos) }


pat_l:
  | Underscore { PWild }
  | c=conid { PCon(c,[]) }
  | c=conid tup=pat { PCon(c,[tup]) }
  | tup=delimited(Lparen, separated_list(Comma,pat), Rparen)
    { mk_tuple (fun x -> PTuple x)
        tup }
  | v=ident { PVar(v) }
  | Tilde { PZero }
  | p = pat Colon t = typ2 { PTyped(p,t) }
  | l = delimited(Lparen, pipe_list, Rparen)
    { fst (List.fold_left (fun x y -> PAlt(x,y), dummy_loc) (List.hd l) (List.tl l) ) }
(* Bad constructors, try again ! *)
pipe_list:
  | p = pat Pipe ps = separated_nonempty_list(Pipe, pat) { p::ps }

clause:
  | p=pat Arrow e=expr { (p,e) }

conid:
  | s=Con { user_id s }

ident:
  | i=Ident { user_id i }

typing:
  | is = ident+ Colon t=typ { List.map (fun i -> (i,t)) is }

oclause:
  | p = pat DoubleArrow q = pat { (p,[],q) }
  | p = pat DoubleArrow q=pat When os = separated_nonempty_list(Comma,typing) { (p,List.concat os,q) }


oclauses:
  | { [] }
  | Pipe c=oclause cs=oclauses { c::cs }

patch_id:
  | Sharp i=IntLit { i }

patch_addr:
  | ps=separated_nonempty_list(Comma,patch_id) { O.PatchFor(ps) }
  | Star { O.DefaultPatch }

patch_ident:
  | i = ident { O.PatchIdent(i) }
  | At i=ident { O.PatchSpecial(i) }

patch:
  | pa=patch_addr RevArrow e=expr { O.PatchHole(pa,e) }
  | Ornament pa=patch_addr RevArrow ps=separated_nonempty_list(Comma, patch_ident) { O.PatchOrnament(pa,ps) }
  | Lifting pa=patch_addr RevArrow ps=separated_nonempty_list(Comma, patch_ident) { O.PatchLifting(pa,ps) }

 
