open Ast
open Id.PrettyNames
(*
module X = Explicit
module I = Implicit
*)
open Ast.Explicit
open Ast.Implicit

let holemap = IdHashtbl.create 13

let maybe_cons do_cons hd tl =
  if do_cons
  then hd::tl
  else tl


let rec pretty_term s : Ast.Explicit.expr -> Ast.Implicit.expr = fun (t,_,l) ->
  (match t with
    | Var(v, _) when IdHashtbl.mem holemap v ->
      Var(UserId(IdHashtbl.find holemap v))
    | Var(v,_) -> Var(id s NsVar v)
    | Let(is_meta,isrec,_,bindings,k) ->
      scope s (fun s' ->
        let slet = if isrec then s' else s in
        Let(is_meta,isrec, List.map (fun (x,_,t) -> (intro s' NsVar x, pretty_term slet t)) bindings,
            pretty_term s' k)
      )
    | Match(e,cs) ->
       Match(pretty_term s e, List.map (fun (p,e) -> scope s (fun s -> (pretty_pat s p, pretty_term s e))) cs)
    | Con(i,_,ts) when is_tuple_name i ->
       Tuple(List.map (pretty_term s) ts)
    | Con(i,_,ts) -> Con(id s NsCons i, List.map (pretty_term s) ts)
    | Abs(x,_,t) -> scope s (fun s -> Abs(intro s NsVar x, pretty_term s t))
    | App(u,v) -> App(pretty_term s u, pretty_term s v)
    | MultiFix(bindings, ret) ->
       scope s
         (fun s ->
          let names = List.map (fun (x,_,_) -> intro s NsVar x) bindings in
          let bindings = List.map2
            (fun n (_,(y,_),e) ->
              (n, scope s (fun s ->  Abs(intro s NsVar y, pretty_term s e), l))
            ) names bindings
          in
          Let(IsShare,true, bindings, (Var(id s NsVar ret), l))
         )
    | ShallowMatch(x, cs) ->
       Match((Var(id s NsVar x), l),
             List.map (fun (c,xs,e) ->
                       scope s
                             (fun s ->
                              ( (if is_tuple_name c
                                 then PTuple(List.map (fun x -> PVar(intro s NsVar x), l) xs), l
                                 else (PCon(id s NsCons c, List.map (fun x -> PVar(intro s NsVar x), l) xs), l)), pretty_term s e))) cs)
    | Mark(NiceName n, e) -> Label(n, pretty_term s e)
    | Mark(m,e) -> fst (pretty_term s e)
    | ShallowCon(c,_,xs,_conid) when is_tuple_name c ->
       Tuple(List.map (fun x -> Var(id s NsVar x),l) xs)
    | ShallowCon(c,_,xs,_conid) -> Con(id s NsCons c, List.map (fun x -> Var(id s NsVar x),l) xs)
  ), l
and pretty_pat s : Ast.Explicit.pat -> Ast.Implicit.pat = fun (p,_,loc) ->
  (match p with
    | PVar(v) -> PVar(intro s NsVar v)
    | PCon(c,_,ps) when is_tuple_name c ->
       PTuple(List.map (pretty_pat s) ps)
    | PCon(c,_,ps) -> PCon(id s NsCons c, List.map (pretty_pat s) ps)
    | PWild -> PWild
    | PZero -> PZero
    | PAlt(p1,p2) -> PAlt(pretty_pat s p1, pretty_pat s p2)
  ), loc


let pretty_top is_printed top_s loc t =
  let top s (TopLet(isrec, bindings)) k =
    scope s (fun s' ->
      let slet = if isrec then s' else s in
      let bindings = List.map (fun (x,_,t) -> (intro_global s' NsVar x, pretty_term slet t)) bindings in
      maybe_cons
        is_printed
        (Ast.StVal(isrec, bindings),loc)
        (k s'))
  in
  top top_s t

let pretty_com is_printed scope loc v k =
  let (Ast.StVal(v),loc)::ds = pretty_top true scope loc v k in
  maybe_cons
    is_printed
    (Ast.StCommented(v), loc)
    ds

let rec pretty_ty s : ty -> user_ty =
  let rec ty s = function
    | TyArrow(t1,t2) -> TyArrow(ty s t1, ty s t2)
    | TyCon(c,ts) when is_tuple_name c -> TyProduct(List.map (ty s) ts)
    | TyCon(c,ts) -> TyCon(id s NsTy c, List.map (ty s) ts)
    | TyVar(x) -> TyVar(id s NsTyVar x)
    | TyProduct(ts) -> TyProduct(List.map (ty s) ts)
  in
  ty s

let pretty_tydef is_printed s (tydef : Explicit.tydef) loc k =
  let Explicit.TyDef(is_global, tydefs) = tydef in
  let is_global = is_printed && is_global in
  let intro_global = if is_global && false then intro_global else intro in
  (*  tn,is_global,targs,tconses) *)
  scope s (fun s ->
    let defs = List.map (fun (tn, targs, tconses) ->
                         scope s (fun s' ->
                                  TyDef(intro_global s NsTy tn,
                                        List.map (intro s' NsTyVar) targs,
                                        List.map (fun (c,tys) ->
                                                  (intro_global s NsCons c,
                                                   List.map (pretty_ty s') tys)) tconses))
                        ) tydefs in
    maybe_cons is_global (Ast.StType(defs), loc) (k s))

let pretty_absdef is_printed s loc x t k =
  let Scheme(vs,t) = t in
  scope s (fun s ->
    scope s (fun s' ->
      maybe_cons
        is_printed
        (Ast.StAbstractVal(intro_global s NsVar x,
                           (ignore (List.map (intro s' NsTyVar) vs); pretty_ty s' t)), loc)
        (k s)
   ))
