
type 'a option =
  | Some of 'a
  | None

type expr =
  | Abs of (expr -> expr)
  | App of expr * expr
  | Const of int

let rec eval e = match e with
  | App (u, v) ->
      begin match eval u with Some (Abs f) -> eval (f v) | _ -> None end
  | _ -> Some(e)

type location = Location of string * int * int
type expr_loc =
  | Abs' of (expr_loc -> expr_loc) * location
  | App' of expr_loc * expr_loc * location
  | Const' of int * location

type ('a, 'err) result =
  | Ok of 'a
  | Error of 'err

type ornament add_loc : expr => expr_loc with
  | Abs f => Abs' (f, _)  when f : add_loc -> add_loc
  | App (u, v) => App' (u, v, _) when u v : add_loc
  | Const i => Const' (i, _)

type ornament ('a, 'err) optres : 'a option => ('a, 'err) result with
  | Some a => Ok a
  | None => Error _

  let eval_id = lifting eval with ornament * <- @id

let eval_incomplete = lifting eval : add_loc -> (add_loc, location) optres

let eval_loc =  lifting eval : add_loc -> (add_loc, location) optres with
  | #4 <- begin match _2 with
            | Ok _ -> (match e with App'(_,_,loc) -> loc)
            | Error x -> x
          end

type expr' =
  | Abs' of (expr' * location -> expr' * location)
  | App' of (expr' * location) * (expr' * location)
  | Const' of int

type ornament add_loc : expr => expr' * location with
  | Abs f => (Abs' f, _) when f : add_loc -> add_loc
  | App (u, v) => (App' (u, v), _) when u v : add_loc
  | Const (i) => (Const' (i), _)

let eval_incomplete = lifting eval : add_loc -> _ optres

let eval' = lifting eval : add_loc -> (_, _) optres with
             | #4 <- begin match _2 with 
                      | Error x -> x 
                      | Ok _ -> (match e with (_, loc) -> loc) (* sourve of the bug *)
                     end
