
type 'a gen = Pair of ('a gen * 'a gen) | Value of 'a | Unit

let gen_map f =
  let rec map x =
    match x with
    | Pair(u,v) -> Pair(map u, map v)
    | Value x -> Value (f x)
    | Unit -> Unit
  in
  map

type 'a list = Nil | Cons of ('a * 'a list)
type 'a goodlist = GNil | GCons of 'a * 'a goodlist

type ornament 'a list_good : 'a list => 'a goodlist with
| Nil => GNil
| Cons (x,xs) => GCons(x, xs) when xs : 'a list_good

type unit = U

type ornament 'a gen_list : 'a gen => 'a list with
| Unit => Nil
| Pair t => Cons t when t : ('a gen_elem * 'a gen_list)
| Value x => ~
and 'a gen_elem : 'a gen => 'a with
| Value x => x
| Unit => ~
| Pair t => ~ when t : ('a gen_elem * 'a gen_list)

let gen_map_unfold f =
  let rec map x =
    match x with
    | Pair(u,v) -> Pair(map_left u, map v)
    | Value x -> Value (f x)
    | Unit -> Unit
  and map_left x =
    match x with
    | Pair(u,v) -> Pair(map_left u, map v)
    | Value x -> Value (f x)
    | Unit -> Unit
  in
  map

let map_list = lifting gen_map_unfold : ('a -> 'b) -> 'a gen_list -> 'b gen_list
let map_goodlist = lifting map_list : ('a -> 'b) -> 'a list_good -> 'b list_good

let rec gen_mapx f = function
  | Pair(u,v) -> Pair(gen_mapx f u, gen_mapx f v)
  | Value x -> Value (f x)
  | Unit -> Unit
(*
let map_list_uf = lifting gen_mapx : ('a -> 'b) -> 'a gen_list -> 'b gen_list
and map_list_uf2 = lifting gen_mapx
*)
  #(mode autolift)
  #(mode info)
  #(mode debug)
let map_list_uf_auto = lifting gen_mapx : ('a -> 'b) -> 'a gen_list -> 'b gen_list

*
type 'a tree = Leaf | Node of ('a * ('a tree * 'a tree))

type ornament 'a gen_tree : 'a gen => 'a tree with
| Unit => Leaf
| Pair t => Node t when t : ('a gen_value * 'a gen_treepair)
| Value x => ~
and 'a gen_value : 'a gen => 'a with
| Value x => x
| Unit => ~
| Pair t => ~ when t : ('a gen_tree * 'a gen_tree)
and 'a gen_treepair : 'a gen => 'a tree * 'a tree with
| Unit => ~
| Pair t => t when t : ('a gen_tree * 'a gen_tree)
| Value x => ~

type 'a goodtree = GLeaf | GNode of 'a * 'a goodtree * 'a goodtree

type ornament 'a tree_good : 'a tree => 'a goodtree with
| Leaf => GLeaf
| Node (x,(l,r)) => GNode(x,l,r) when l r : 'a tree_good

let gen_map_unfold3 f =
  let rec map_main x =
    match x with
    | Pair(u,v) -> Pair(map_value u, map_pair v)
    | Value x -> Value (f x)
    | Unit -> Unit
  and map_pair x =
    match x with
    | Pair(u,v) -> Pair(map_main u, map_main v)
    | Value x -> Value(f x)
    | Unit -> Unit
  and map_value x =
    match x with
    | Pair(u,v) -> Pair(map_main u, map_main v)
    | Value x -> Value (f x)
    | Unit -> Unit
  in
  map_main

let map_tree = lifting gen_map_unfold3 : ('a -> 'b) -> 'a gen_tree -> 'b gen_tree
let map_goodtree = lifting map_tree : ('a -> 'b) -> 'a tree_good -> 'b tree_good

(* Can we merge the two ornaments into one ? *)
type ornament 'a gen_goodlist : 'a gen => 'a goodlist with
| Unit => GNil
| Pair (u,v) => GCons(u,v) when u : 'a gen_glval, v : 'a gen_goodlist
| Value x => ~
and 'a gen_glval : 'a gen => 'a with
| Value x => x
| Pair t => ~ when t : ('a gen_glval * 'a gen_goodlist)
| Unit => ~

let map2_goodlist = lifting gen_map_unfold : ('a -> 'b) -> 'a gen_goodlist -> 'b gen_goodlist

(*
(* The following does not work: *)
type ornament 'a ggen_tree : 'a gen => 'a goodtree with
| Unit => GLeaf
| Pair(x,Pair(l,r)) => GNode(x,l,r) when x : 'a ggen_value, l r : 'a ggen_tree
| Value x => ~
and 'a ggen_value : 'a gen => 'a with
| Value x => x
| Unit => ~
| Pair t => ~ when t : ('a ggen_tree * 'a ggen_tree) 
*)
