open Options

let first_some = function
  | Some _ as x -> fun _ -> x
  | None -> fun x -> x

let bind_option f = function
  | Some x -> f x
  | None -> None

let map_option f = function
  | Some x -> Some (f x)
  | None -> None

let rec map_filter f = function
  | [] -> []
  | x::xs -> match f x with
      | None -> map_filter f xs
      | Some y -> y::map_filter f xs

let rec search = function
  | [] -> fun f -> raise Not_found
  | x::xs -> fun f -> match f x with
      | Some(y) -> y
      | None -> search xs f

let rec find l f = search l (fun x -> if f x then Some(x) else None)

let try_find l f = try Some(find l f) with Not_found -> None

let try_search l f = try Some(search l f) with Not_found -> None

let find_duplicate l =
  let rec scan = function
    | [] -> None
    | x::xs -> if List.exists (fun x' -> x = x') xs then Some(x) else scan xs
  in
  scan l

let debug eta =
  if !debug_mode
  then Format.eprintf (eta ^^ "@.")
  else Format.ifprintf Format.err_formatter eta

let say eta =
  if !display_info
  then Format.eprintf (eta ^^ "@.")
  else Format.ifprintf Format.err_formatter eta

let warn eta =
  Format.eprintf ("Warning: "^^eta^^"@.")

type ('a,'b) choice = Inl of 'a | Inr of 'b
let fst3 (x,_,_) = x
let snd3 (_,x,_) = x
let thr3 (_,_,x) = x

let curfile = ref ""

let list_init n f = Array.to_list (Array.init n f)

let rec cut_list n l =
  match n, l with
  | 0, (x::xs) -> ([], x, xs)
  | n, (x::xs) when n > 0 ->
     let (ls,a,rs) = cut_list (n-1) xs in
     (x::ls,a,rs)
  | _, [] -> failwith "cut_list: empty list"
  | _ -> failwith "cut_list: negative argument"

let rec modify_one f = function
  | [] -> None
  | x::xs ->
     begin match f x with
     | None -> begin match modify_one f xs with
       | None -> None
       | Some xs -> Some(x::xs)
     end
     | Some x -> Some(x::xs)
     end
