open Str

let file_to_string filename = 
  let fd = open_in filename in
  let rec input all = 
    try input (input_line fd :: all)
    with End_of_file -> close_in fd; all 
  in
  let all = input [] in
  String.concat "\n" (List.rev all)

let begin_program =
  (* Str.regexp "^ *\\\\begin{program}\\(\\[[^]]*\\]\\)?\\({[^}]*}\\)?" *)
  Str.regexp "^ *\\\\begin{program}\\(\\[[^]]*\\]\\)?\\({[^}]*}\\)?"
let end_program = Str.regexp "^ *\\\\end{program}"

exception Unclosed of int * int
let rec process_string str iput oput =
  let rec process pos =
    try 
      let pos = search_forward begin_program str pos in
      let matched_group n str =
        try matched_group n str with Not_found -> "" in
      let bracket = matched_group 1 str in
      let braces = matched_group 2 str in
      let pos_begin = match_end () in
      let pos_end =  
        try search_forward end_program str pos_begin
        with Not_found -> raise (Unclosed (pos, pos_begin))
      in
      let pos = match_end () in
      let region out =
        output_substring out str pos_begin (pos_end - pos_begin) in
      let () =
        match bracket, braces with
        | "[input]", _ | _, "{input}" -> region iput
        | "[output]", _ | _, "{output}" -> region oput
        | "[inout]", _ | _, "{inout}" -> region iput; region oput
        | _ -> () in
      process pos
    with Not_found -> ()
in
process 0 


let process src =
  let basename = Filename.chop_extension src in
  let texname = basename ^ ".tex" in
  if not (src = texname) then failwith (src ^ "does not end with .tex");
  Printf.eprintf "Processing %s\n%!" texname;
  let text = file_to_string texname in
  let iput = open_out (basename ^ ".ml") in
  let oput = open_out (basename ^ ".ref.ml") in
  let () =
    try process_string text iput oput
    with Unclosed (b,e) -> 
      Printf.printf "Unbalance (%s, %d, %d)" src b e
  in
  close_out iput;
  close_out oput;
  ()
    
let main () = 
  let files = List.tl (Array.to_list Sys.argv) in
  List.iter process files

let () = main()
